import datetime as dt
from logging import getLogger, WARNING, INFO, DEBUG
from unittest import TestCase

from ch2 import PROGNAME
from ch2.common.args import mm
from ch2.common.io import data_hash
from ch2.common.log import configure_log
from ch2.commands.args import make_parser, NamespaceWithVariables, DB_VERSION
from ch2.common.names import USER
from ch2.common.user import make_user_database
from ch2.sql.config import Config
from ch2.sql.support import Base

log = getLogger(__name__)


class LogTestCase(TestCase):

    def setUp(self):
        configure_log(PROGNAME, '/tmp/ch2-test.log', verbosity=5, levels={
            'sqlalchemy': WARNING,
            'matplotlib': INFO,
            'bokeh': DEBUG,
            'tornado': INFO,
            'sentinelsat': DEBUG,
            'werkzeug': DEBUG,
            'ch2': DEBUG,
            '__main__': DEBUG
        })


def random_test_user(args=(mm(USER), 'postgres')):
    parser = make_parser()
    ns = NamespaceWithVariables._from_ns(parser.parse_args(args=args), PROGNAME, DB_VERSION)
    config = Config(ns)
    user = data_hash(str(dt.datetime.now()))[:6]
    log.info(f'User/database {user}')
    user_config = make_user_database(config, user, '')
    log.info('Creating tables')
    Base.metadata.create_all(user_config.db.engine)
    return user



from logging import getLogger

import numpy as np
import pandas as pd
from scipy.signal import find_peaks

from .utils import ProcessCalculator, IntervalCalculatorMixin
from ..pipeline import OwnerInMixin, LoaderMixin
from ...data import Statistics
from ...lib import format_date, local_date_to_time
from ...names import Titles as T, Summaries as S, Units as U, Names as N
from ...sql import StatisticJournalInteger, StatisticJournalType

log = getLogger(__name__)


class RestHRCalculator(LoaderMixin, OwnerInMixin, IntervalCalculatorMixin, ProcessCalculator):
    '''
    Used to calculate rest HR from quartiles, but it was never clear we had *the* rest value rather
    than some general lower value.

    This way - by looking for the first peak that's not noise - we are finding something that perhaps
    is more meaningful.  It's a low heart rate that you spent a fair amount of time at.
    '''

    def __init__(self, *args, schedule='d', **kargs):
        # permanent blocks clearing of dirty values - cleared explicitly by
        super().__init__(*args, schedule=schedule, permanent=True, **kargs)

    def _startup(self, s):
        super()._startup(s)
        self._provides(s, T.REST_HR, StatisticJournalType.INTEGER, U.BPM, S.join(S.MIN, S.MSR),
                       'The rest heart rate.')

    def _read_data(self, s, interval):
        return Statistics(s, start=local_date_to_time(interval.start),
                          finish=local_date_to_time(interval.finish)). \
            by_name(self.owner_in, N.HEART_RATE).df

    def _calculate_results(self, s, interval, df, loader):
        if not df.empty:
            hist = pd.cut(df[N.HEART_RATE], np.arange(30, 90), right=False).value_counts(sort=False)
            peaks, _ = find_peaks(hist)
            for peak in peaks:
                rest_hr = hist.index[peak].left
                measurements = hist.loc[rest_hr]
                if measurements > len(df) * 0.01:
                    log.debug(f'Rest HR for {format_date(interval.start)} is {rest_hr} with {measurements} values')
                    # conversion to int as value above is numpy int64
                    loader.add_data(N.REST_HR, interval, int(rest_hr), local_date_to_time(interval.start))
                    return
                else:
                    log.debug(f'Skipping rest HR at {format_date(interval.start)} because too few measurements '
                              f'({measurements}/{len(df)})')
        log.warning(f'Unable to calculate rest HR at {format_date(interval.start)}')





from ..database import add_process
from ..profile import Profile
from ...pipeline.read.activity import ActivityReader
from ...pipeline.read.garmin import GarminReader
from ...pipeline.read.monitor import MonitorReader
from ...sql.types import short_cls


def garmin(config):
    '''
## garmin

This extends the default configuration with download of monitor data from Garmin.

It requires the user and password to be defined as constants.
    '''
    Garmin(config).load()


class Garmin(Profile):

    def _load_read_pipeline(self, s):
        sport_to_activity = self._sport_to_activity()
        record_to_db = self._record_to_db()
        add_process(s, ActivityReader, owner_out=short_cls(ActivityReader),
                    sport_to_activity=sport_to_activity, record_to_db=record_to_db)
        monitor_reader = add_process(s, MonitorReader)
        # add these, chained so that we load available data (know what is missing),
        # download new data, and load new data
        garmin_reader = add_process(s, GarminReader, blocked_by=[monitor_reader],
                                    force_all=True, max_days=365)
        add_process(s, MonitorReader, blocked_by=[garmin_reader])




# these are the namespaces (owners) of statistics
from .calculate import *
from .read.activity import ActivityReader
from .read.monitor import MonitorReader
from .read.garmin import GarminReader
from ..sql import ActivityTopic, DiaryTopic




import datetime as dt
from collections import defaultdict
from logging import getLogger

import numpy as np
from sqlalchemy import desc, and_, func
from sqlalchemy.orm import aliased

from .utils import AbortImportButMarkScanned, ProcessFitReader
from ..calculate import RestHRCalculator
from ..loader import Loader
from ..pipeline import LoaderMixin
from ...common.date import time_to_local_date, format_time, to_time, dates_from, now
from ...data.frame import read_query
from ...fit.format.records import fix_degrees, unpack_single_bytes, merge_duplicates
from ...fit.profile.profile import read_fit
from ...names import N, T, U
from ...sql import MonitorJournal, StatisticJournalInteger, StatisticName, StatisticJournal, Interval
from ...sql.database import StatisticJournalType, Source
from ...sql.utils import add

log = getLogger(__name__)

ACTIVITY_TYPE_ATTR = 'activity_type'
HEART_RATE_ATTR = 'heart_rate'
MONITORING_ATTR = 'monitoring'
MONITORING_INFO_ATTR = 'monitoring_info'
STEPS_ATTR = 'steps'


def missing_dates(s, force=False, max_days=30):
    # we don't try to be perfect here.  the idea is that it's called to get the latest
    # updates, rather than fill in all gaps (do the bulk download thing for that).
    # we also don't try to get fractional data.
    # and as for timezones... we just assume garmin uses the local timezone.
    latest = s.query(MonitorJournal).order_by(desc(MonitorJournal.start)).limit(1).one_or_none()
    if latest is None:
        log.warning('No existing monitor data - ' +
                    'do a bulk download instead: https://www.garmin.com/en-US/account/datamanagement/')
        return
    # find the mid-point to avoid any problems with timezones and edge cases
    # (these files don't span more than a day)
    seconds = (latest.finish - latest.start).total_seconds() / 2
    start = time_to_local_date(latest.start + dt.timedelta(seconds=seconds))
    days = (time_to_local_date(now()) - start).days
    if days > max_days and not force:
        raise Exception('Too many days (%d) - ' % days +
                        'do a bulk download instead: https://www.garmin.com/en-US/account/datamanagement/')
    if days:
        yield from dates_from(start)
    else:
        log.warning('No dates to download')


NEW_STEPS = N._new(N.STEPS)
STEPS_DESCRIPTION = '''The increment in steps read from the FIT file.'''


class MonitorReader(LoaderMixin, ProcessFitReader):

    '''
    These overlap (well, often one starts when another ends),
    so we read everything in and then, at the end, remove overlaps.
    '''

    def __init__(self, config, *args, **kargs):
        from ...commands.upload import MONITOR
        super().__init__(config, *args, sub_dir=MONITOR, **kargs)

    def _startup(self, s):
        super()._startup(s)
        self._provides(s, T.HEART_RATE, StatisticJournalType.INTEGER, U.BPM, None,
                       'The instantaneous heart rate.')
        self._provides(s, T.CUMULATIVE_STEPS, StatisticJournalType.INTEGER, U.STEPS_UNITS, None,
                       'The number of steps in a day to this point in time.')
        self._provides(s, T.STEPS, StatisticJournalType.INTEGER, U.STEPS_UNITS, None,
                       'The increment in steps read from the FIT file.')

    def _get_loader(self, s, **kargs):
        return super()._get_loader(s, cls=MonitorLoader, **kargs)

    def _missing(self, s):
        missing = super()._missing(s)
        if missing:
            # clear rest heart rate so it will be recalculated
            Interval.clean(s, owner=RestHRCalculator)
        return missing

    @staticmethod
    def parse_records(data):
        return MonitorReader.read_fit_file(data, merge_duplicates, fix_degrees, unpack_single_bytes)

    @staticmethod
    def read_first_timestamp(path, records):
        return MonitorReader._first(path, records, MONITORING_INFO_ATTR).value.timestamp

    @staticmethod
    def read_last_timestamp(path, records):
        return MonitorReader._last(path, records, MONITORING_ATTR).value.timestamp

    def _read_data(self, s, file_scan):
        records = self.parse_records(read_fit(file_scan.path))
        first_timestamp = self.read_first_timestamp(file_scan.path, records)
        last_timestamp = self.read_last_timestamp(file_scan.path, records)
        if first_timestamp == last_timestamp:
            log.warning('File %s is empty (no timespan)' % file_scan)
            raise AbortImportButMarkScanned()
        if not first_timestamp:
            raise Exception('Missing timestamp in %s' % file_scan)

        log.info(f'Importing monitor data from {file_scan} '
                 f'for {format_time(first_timestamp)} - {format_time(last_timestamp)}')
        if s.query(MonitorJournal).filter(MonitorJournal.file_hash == file_scan.file_hash).count():
            raise Exception(f'Duplicate for {file_scan.path}')  # should never happen
        mjournal = add(s, MonitorJournal(start=first_timestamp, finish=last_timestamp,
                                         file_hash_id=file_scan.file_hash.id))
        return mjournal, (first_timestamp, last_timestamp, mjournal, records)

    def _load_data(self, s, loader, data):
        first_timestamp, last_timestamp, mjournal, records = data
        steps_by_activity = defaultdict(lambda: 0)
        for record in records:
            if HEART_RATE_ATTR in record.data and record.data[HEART_RATE_ATTR][0][0]:
                loader.add_data(N.HEART_RATE, mjournal, record.data[HEART_RATE_ATTR][0][0], record.timestamp)
            if STEPS_ATTR in record.data:
                # we ignore activity type here (used to store it when activity group and statistic name
                # were mixed together, but never used it anywhere)
                reset = False
                for activity, steps in zip(record.data[ACTIVITY_TYPE_ATTR][0], record.data[STEPS_ATTR][0]):
                    reset = reset or steps < steps_by_activity[activity]
                if reset: steps_by_activity = defaultdict(lambda: 0)
                for activity, steps in zip(record.data[ACTIVITY_TYPE_ATTR][0], record.data[STEPS_ATTR][0]):
                    steps_by_activity[activity] = steps
                total = sum(steps_by_activity.values())
                loader.add_data(N.CUMULATIVE_STEPS, mjournal, total, record.timestamp)

    def _shutdown(self, s):
        super()._shutdown(s)
        if not self.worker:
            log.info('Calculating differential in main thread')
            self._fix_overlapping_monitors(s)
            s.commit()
            self._update_differential(s)

    def _fix_overlapping_monitors(self, s):
        # this used to be done in python and was very slow
        self._delete_nested(s)
        self._shift_finish(s)
        self._delete_past_finish(s)

    def _delete_nested(self, s):
        # first, remove any monitor journal entries which are completely enclosed by others
        m1 = aliased(MonitorJournal)
        m2 = aliased(MonitorJournal)
        nested = s.query(m2.id).join(m1). \
            filter(m1.start <= m2.start,
                   m1.finish >= m2.finish,
                   m1.id != m2.id)
        q = s.query(Source).filter(Source.id.in_(nested))
        count = q.count()
        if count:
            log.warning(f'Deleting {count} nested journal entries')
            q.delete()
        else:
            log.debug('No nested journal entries')

    def _shift_finish(self, s):
        # next, fix up overlapping monitor journal entries so that they abut
        log.info(f'Adjusting finish')
        m1 = aliased(MonitorJournal)
        m2 = aliased(MonitorJournal)
        # this is giving a warning but post to sqlalchemy group was deleted
        overlap = s.query(m1.id.label('id'), m2.start.label('finish')). \
            filter(m1.start < m2.start,
                   m1.finish > m2.start,
                   m1.id != m2.id).cte()
        s.query(MonitorJournal). \
            filter(MonitorJournal.id == overlap.c.id). \
            update({MonitorJournal.finish: overlap.c.finish})

    def _delete_past_finish(self, s):
        # finally, drop any values that are not included in the shortened journals
        q = s.query(StatisticJournal.id). \
            join(MonitorJournal). \
            filter(StatisticJournal.source_id == MonitorJournal.id,
                   StatisticJournal.time >= MonitorJournal.finish)
        count = q.count()
        if count:
            log.warning(f'Deleting {count} orphan statistics')
            # dance around the inability to call q.delete() directly
            s.query(StatisticJournal).filter(StatisticJournal.id.in_(q)).delete(synchronize_session=False)
        else:
            log.debug('No orphan statistics')

    def _update_differential(self, s):
        # this reads CUMULATIVE_STEPS (which is what was in the files) and any existing STEPS
        # then calculates what STEPS should be and fixes up any incorrect or missing data
        # (i guess maybe slightly faster when running incrementally)
        df = self._read_diff(s)
        df = self._calculate_diff(df)
        self._write_diff(s, df)

    def _read_diff(self, s):
        qs = s.query(StatisticJournalInteger.time.label(N.TIME),
                     StatisticJournalInteger.value.label(N.STEPS)). \
            join(StatisticName). \
            filter(StatisticName.name == N.STEPS,
                   StatisticName.owner == self.owner_out).cte()
        q = s.query(StatisticJournalInteger.time.label(N.TIME),
                    StatisticJournalInteger.source_id.label(N.SOURCE),
                    StatisticJournalInteger.value.label(N.CUMULATIVE_STEPS),
                    qs.c.steps.label(N.STEPS)). \
            join(StatisticName).outerjoin(qs, StatisticJournalInteger.time == qs.c.time). \
            filter(StatisticName.name == N.CUMULATIVE_STEPS,
                   StatisticName.owner == self.owner_out). \
            order_by(StatisticJournalInteger.time)
        # log.debug(q)
        df = read_query(q, index=N.TIME)
        return df

    def _calculate_diff(self, df):
        df[NEW_STEPS] = df[N.CUMULATIVE_STEPS].diff()
        df.loc[df[NEW_STEPS] < 0, NEW_STEPS] = df[N.CUMULATIVE_STEPS]
        df.loc[df[NEW_STEPS].isna(), NEW_STEPS] = df[N.CUMULATIVE_STEPS]
        return df

    def _write_diff(self, s, df):
        steps = StatisticName.from_name(s, N.STEPS, self.owner_out)
        times = df.loc[(df[NEW_STEPS] != df[N.STEPS]) & ~df[N.STEPS].isna()].index.astype(np.int64) / 1e9
        if len(times):
            times = [to_time(time) for time in times]
            n = s.query(func.count(StatisticJournal.id)). \
                filter(StatisticJournal.time.in_(times),
                       StatisticJournal.statistic_name == steps).scalar()
            log.warning(f'Deleting {n} {N.STEPS} entries')
            s.query(StatisticJournal.id). \
                filter(StatisticJournal.time.in_(times),
                       StatisticJournal.statistic_name == steps).delete(synchronize_session=False)
        loader = self._get_loader(s, owner=self.owner_out, add_serial=False)
        for time, row in df.loc[(df[NEW_STEPS] != df[N.STEPS]) & ~df[NEW_STEPS].isna()].iterrows():
            loader.add_data(N.STEPS, row[N.SOURCE], int(row[NEW_STEPS]), time)
        loader.load()


class MonitorLoader(Loader):

    def _resolve_duplicate(self, name, instance, prev):
        log.warning(f'Using max of duplicate values at {instance.time} for {name} '
                    f'({instance.value}/{prev.value})')
        prev.value = max(prev.value, instance.value)




import datetime as dt
from logging import getLogger

from sqlalchemy.sql import func
from sqlalchemy.sql.functions import count

from .utils import ProcessCalculator
from ..pipeline import LoaderMixin, OwnerInMixin
from ...common.date import format_dateq
from ...common.log import log_current_exception
from ...lib import local_date_to_time, time_to_local_date, to_date
from ...names import T, N, U, S
from ...sql import MonitorJournal, StatisticJournalInteger, StatisticName, StatisticJournal, Composite, \
    CompositeComponent, Source, StatisticJournalType
from ...sql.utils import add

log = getLogger(__name__)

QUARTER_DAY = 6 * 60 * 60


class StepsCalculator(LoaderMixin, OwnerInMixin, ProcessCalculator):
    '''
    This is a little unusual, in that we can calculate results from partial data and then, when we have
    more data, we need to delete the previous values.  So we need to be careful (1) in deciding when
    we have new data; (2) in avoiding duplicates; and (3) in tracking composite sources.
    '''

    def _startup(self, s):
        super()._startup(s)
        self._provides(s, T.DAILY_STEPS, StatisticJournalType.INTEGER, U.STEPS_UNITS,
                       S.join(S.SUM, S.AVG, S.CNT, S.MAX, S.MSR),
                       'The number of steps in a day.')

    def _missing(self, s):
        # any day that has an unused monitor journal is a missing day
        Composite.clean(s)
        monitor_stats = s.query(StatisticJournal.id). \
            join(CompositeComponent, CompositeComponent.output_source_id == StatisticJournal.source_id). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.owner == self.owner_out,
                   CompositeComponent.input_source_id == MonitorJournal.id)
        unused_sources = s.query(MonitorJournal).filter(~monitor_stats.exists())
        log.debug(unused_sources)
        dates = set()
        for source in unused_sources.all():
            for date in self._dates_for_source(source):
                dates.add(format_dateq(date))
        missing = sorted(dates)
        if missing:
            log.debug(f'Missing: {missing[0]} - {missing[-1]} / {len(missing)}')
        return missing

    @staticmethod
    def _dates_for_source(mjournal):
        start = time_to_local_date(mjournal.start)
        finish = time_to_local_date(mjournal.finish) + dt.timedelta(days=1)
        while start <= finish:
            yield start
            start += dt.timedelta(days=1)

    def _delete_time_range(self, s, start=None, finish=None):
        composite_ids = s.query(Composite.id). \
            join(StatisticJournal, Composite.id == StatisticJournal.source_id). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.owner == self.owner_out)
        if start:
            composite_ids = composite_ids.filter(StatisticJournal.time >= start)
        if finish:
            composite_ids = composite_ids.filter(StatisticJournal.time <= finish)
        log.debug(f'Delete query: {composite_ids}')
        q = s.query(Source).filter(Source.id.in_(composite_ids))
        count = q.count()
        if count:
            log.warning(f'Deleting {count} Composite sources ({start} - {finish})')
            q.delete(synchronize_session=False)
            s.commit()

    def _run_one(self, missed):
        # careful here to make summertime work correctly
        finish = local_date_to_time(to_date(missed) + dt.timedelta(days=1))
        start = local_date_to_time(to_date(missed))
        with self._config.db.session_context() as s:
            # delete any previous, incomplete data
            if s.query(count(Composite.id)). \
                    join(StatisticJournal, Composite.id == StatisticJournal.source_id). \
                    join(StatisticName). \
                    filter(StatisticJournal.time >= start,
                           StatisticJournal.time < finish,
                           StatisticName.owner == self.owner_out).scalar():
                self._delete_time_range(s, start, finish)
            try:
                input_source_ids, daily_steps = self._read_data(s, start, finish)
                if not input_source_ids: raise Exception(f'No sources found on {start}')
                if daily_steps is None: raise Exception(f'No steps data for {start}')
                output_source = add(s, Composite(n_components=len(input_source_ids)))
                for input_source_id in input_source_ids:
                    s.add(CompositeComponent(input_source_id=input_source_id, output_source=output_source))
                s.commit()
                loader = self._get_loader(s, add_serial=False, clear_timestamp=False)
                loader.add_data(N.DAILY_STEPS, output_source, daily_steps, start)
                loader.load()
                self._prev_loader = loader
            except Exception as e:
                log.error(f'No statistics for {start} - {finish}: ({e})')
                log_current_exception()

    def _read_data(self, s, start, finish):
        daily_steps = s.query(func.sum(StatisticJournalInteger.value)).join(StatisticName). \
            filter(StatisticName.name == N.STEPS,
                   StatisticName.owner == self.owner_in,
                   StatisticJournalInteger.time < finish,
                   StatisticJournalInteger.time >= start).scalar()
        input_source_ids = [row[0] for row in s.query(MonitorJournal.id).
            filter(MonitorJournal.start < finish,
                   MonitorJournal.finish >= start).all()]
        return input_source_ids, daily_steps



from logging import getLogger
from time import sleep

from requests import HTTPError

from .monitor import missing_dates
from ..pipeline import ProcessPipeline
from ...commands.args import DATA
from ...common.date import now, local_time_to_time, format_date, to_date, dates_from, time_to_local_time, format_dateq
from ...common.log import log_current_exception
from ...fit.download.connect import GarminConnect
from ...sql import Constant, SystemConstant

log = getLogger(__name__)

GARMIN_USER = 'garmin_user'
GARMIN_PASSWORD = 'garmin_password'


class GarminReader(ProcessPipeline):

    def __init__(self, *args, force_all=False, max_days=100, **kargs):
        self.__force_all = force_all
        self.__max_days = max_days
        self.__user = None
        self.__password = None
        super().__init__(*args, **kargs)

    def _startup(self, s):
        self.__user = Constant.get_single(s, GARMIN_USER, none=True)
        self.__password = Constant.get_single(s, GARMIN_PASSWORD, none=True)
        super()._startup(s)

    def _missing(self, s):
        last = self._config.get_constant(SystemConstant.LAST_GARMIN, none=True)
        if not (self.__user and self.__password):
            log.warning('No username or password defined for Garmin download')
        elif last and (now() - local_time_to_time(last)).total_seconds() < 12 * 60 * 60:
            log.warning(f'Too soon since previous call ({last}; 12 hours minimum)')
        else:
            try:
                dates = [format_dateq(date)
                         for date in missing_dates(s, force=self.__force_all, max_days=self.__max_days)]
                if dates:
                    log.debug(f'Download Garmin from {dates[0]}')
                    return dates[:1]  # dates[0] in a list
                else:
                    log.debug('No Garmin data to download')
            except Exception as e:
                log_current_exception()
                log.warning(e)

    def _run_one(self, missed):
        # is only called once with the start date
        data_dir = self._config.args._format_path(DATA)
        connect = GarminConnect(log_response=False)
        connect.login(self.__user, self.__password)

        for repeat, date in enumerate(dates_from(to_date(missed))):
            if repeat:
                sleep(1)
            log.info('Downloading data for %s' % date)
            try:
                connect.get_monitoring_to_fit_file(date, data_dir)
            except HTTPError:
                log_current_exception(traceback=False)
                if self.__force_all:
                    log.warning(f'No data for {date}, but continuing')
                else:
                    log.info('End of data')
                    break

        self._config.set_constant(SystemConstant.LAST_GARMIN, time_to_local_time(now()), True)
        return



from logging import getLogger
from tempfile import TemporaryDirectory

import sqlalchemy.sql.functions as func

from ch2.commands.args import V, DEV, MONITOR, bootstrap_db, BASE, UPLOAD
from ch2.commands.upload import upload
from ch2.common.args import mm, m
from ch2.common.date import to_time, local_date_to_time
from ch2.config.profiles.default import default
from ch2.names import N
from ch2.pipeline.calculate.steps import StepsCalculator
from ch2.sql.tables.monitor import MonitorJournal
from ch2.sql.tables.statistic import StatisticJournal, StatisticName
from tests import LogTestCase, random_test_user

log = getLogger(__name__)


class TestMonitor(LogTestCase):

    def test_monitor(self):
        user = random_test_user()
        bootstrap_db(user, m(V), '5')
        with TemporaryDirectory() as f:
            bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), configurator=default)
            config = bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), UPLOAD,
                                  'data/test/source/personal/25822184777.fit')
            upload(config)
            with config.db.session_context() as s:
                n = s.query(func.count(StatisticJournal.id)).scalar()
                self.assertEqual(134, n)
                mjournal = s.query(MonitorJournal).one()
                self.assertNotEqual(mjournal.start, mjournal.finish)

    def test_values(self):
        user = random_test_user()
        with TemporaryDirectory() as f:
            try:
                bootstrap_db(user, m(V), '5')
                bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), configurator=default)
                for file in ('24696157869', '24696160481', '24696163486'):
                    config = bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), UPLOAD,
                                               'data/test/source/personal/andrew@acooke.org_%s.fit' % file)
                    upload(config)
                with config.db.session_context() as s:
                    mjournals = s.query(MonitorJournal).order_by(MonitorJournal.start).all()
                    assert mjournals[2].start == to_time('2018-09-06 15:06:00'), mjournals[2].start
                    # steps
                    summary = s.query(StatisticJournal).join(StatisticName). \
                        filter(StatisticJournal.time >= local_date_to_time('2018-09-06'),
                               StatisticJournal.time < local_date_to_time('2018-09-07'),
                               StatisticName.owner == StepsCalculator,
                               StatisticName.name == N.DAILY_STEPS).one()
                    # connect has 12757 for this date,
                    self.assertEqual(summary.value, 12757)
            except Exception as e:
                raise  # space for debugger to halt

    FILES = ('25505915679', '25519562859', '25519565531', '25532154264', '25539076032', '25542112328')

    def generic_bug(self, files, join=False):
        user = random_test_user()
        with TemporaryDirectory() as f:
            try:
                config = bootstrap_db(user, m(V), '5')
                bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), configurator=default)
                if join:
                    files = ['data/test/source/personal/andrew@acooke.org_%s.fit' % file for file in files]
                    config = bootstrap_db(user, mm(BASE), f, mm(DEV), UPLOAD, *files)
                    upload(config)
                else:
                    for file in files:
                        config = bootstrap_db(user, mm(BASE), f, mm(DEV), UPLOAD,
                                              'data/test/source/personal/andrew@acooke.org_%s.fit' % file)
                        upload(config)
                with config.db.session_context() as s:
                    # steps
                    summary = s.query(StatisticJournal).join(StatisticName). \
                        filter(StatisticJournal.time >= local_date_to_time('2018-10-07'),
                               StatisticJournal.time < local_date_to_time('2018-10-08'),
                               StatisticName.owner == StepsCalculator,
                               StatisticName.name == N.DAILY_STEPS).one()
                    # connect has 3031 for this date.
                    self.assertEqual(summary.value, 3031)
            except Exception as e:
                raise  # space for debugger to halt

    def test_bug(self):
        self.generic_bug(sorted(self.FILES))

    def test_bug_reversed(self):
        self.generic_bug(sorted(self.FILES, reverse=True))

    def test_bug_join(self):
        self.generic_bug(sorted(self.FILES), join=True)

    def test_bug_reversed_join(self):
        self.generic_bug(sorted(self.FILES, reverse=True), join=True)

    # issue 6
    def test_empty_data(self):
        user = random_test_user()
        with TemporaryDirectory() as f:
            bootstrap_db(user, mm(BASE), f, m(V), '5')
            bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV), configurator=default)
            config = bootstrap_db(user, mm(BASE), f, m(V), '5', mm(DEV),
                                      UPLOAD, 'data/test/source/other/37140810636.fit')
            upload(config)
            with config.db.session_context() as s:
                n = s.query(func.count(StatisticJournal.id)).scalar()
                self.assertEqual(12, n)
                mjournal = s.query(MonitorJournal).one()
                self.assertNotEqual(mjournal.start, mjournal.finish)
